
pybotnet_version = '1.0.0'
pybotnet_github_link = 'https://github.com/onionj/pybotnet'
pybotnet_license = 'GNU Lesser General Public License v3 or later (LGPLv3+)'
pybotnet_pypi_link = 'https://pypi.org/project/pybotnet/'


pybotnet_info = f"""
            ************ PyBotNet ************
version:       {pybotnet_version}         
licence:       {pybotnet_license}                        
github link:   {pybotnet_github_link}
pypi link:     {pybotnet_pypi_link}

            ************           ************
"""
